/*
 * Copyright (C) 2005 Andre Noll <noll@mathematik.tu-darmstadt.de>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "audiod.cmdline.h"
#include "para.h"

#define MAX_STAT_CLIENTS 50
#define MAX_PENALTY 20
#define MAX_PENALTY_MS 2000
#define PENALTY_RESET_TIME 10
#define PENALTY_INCREASE_TIME 2

extern const char *status_item_list[NUM_STAT_ITEMS];
static int stat_pipe = -1, signal_pipe[2];
struct gengetopt_args_info conf;
static int offset_seconds, length_seconds, sa_time_diff_sign = 1;
static struct timeval server_stream_start, sa_time_diff;
static int current_decoder = -1, num_stat_clients;

#define STREAM_IO_BUFSIZE 4096

struct filter {
	char *buf;
	int loaded, killed;
	pid_t pid;
	int read_fd, write_fd;
	int rcheck, wcheck;
};

struct audio_format_info {
	char *name;
	char *rcmd, *wcmd;
	unsigned int bufsize;
	unsigned long penalty;
	/* last start time of the decoder for that format in any slot */
	struct timeval stime;
	char **filter_cmds;
	unsigned int num_filters;
};


static struct audio_format_info afi[] = { /* audio format info */
#if 1
	{
		.name = "mp3",
		.rcmd = "para_recv",
		.wcmd = "madplay -",
		.bufsize = 4096,
	},
#endif
#ifdef HAVE_OGGVORBIS
	{
		.name = "ogg",
		.rcmd = "para_recv --header",
		.wcmd = "ogg123 -",
		.bufsize = 4096,
	},
#endif
	{
		.name = NULL,
	}
};

struct stream_io {
	int format, read_fd, write_fd;
	pid_t rpid, wpid;
	/* last read/write time, last start time od reader/writer */
	struct timeval rtime, wtime, rstime, wstime;
	int buflen;
	char *buf;
	int rcheck, wcheck;
	int rkilled, wkilled;
	struct filter *filters;
};

struct command {
	char *name;
	int (*handler)(int, int, char**);
	char *description;
	char *usage;
	char *help;
};
int com_cycle(int, int, char **);
int com_help(int, int, char **);
int com_off(int, int, char **);
int com_on(int, int, char **);
int com_sb(int, int, char **);
int com_stat(int, int, char **);
int com_term(int, int, char **);
struct command cmds[] = {
	{
		.name = "cycle",
		.handler = com_cycle,
		.description = "switch to next mode",
		.usage = "cycle",
		.help = "on -> standby -> off -> on\n"
	},
	{
		.name = "help",
		.handler = com_help,
		.description = "display command list or help for given command",
		.usage = "help [command]",
		.help =
                "When I was younger, so much younger than today, I   \n"
                "never needed anybody's help in any way.  But now    \n"
                "these days are gone, I'm not so self assured. Now I \n"
                "find I've changed my mind and opened up the doors.  \n"
                "-- Beatles: Help                                    \n"
	},
	{
		.name = "off",
		.handler = com_off,
		.description = "deactivate para_audiod",
		.usage = "off",
		.help =
                "Close connection to para_server and stop all        \n"
                "decoders.                                           \n"
	},
	{
		.name = "on",
		.handler = com_on,
		.description = "activate para_audiod",
		.usage = "on",
		.help =
                "Establish connection to para_server, retrieve       \n"
                "para_server's current status. If playing, start     \n"
                "corresponding decoder. Otherwise stop all decoders. \n"
	},
	{
		.name = "sb",
		.handler = com_sb,
		.description = "enter standby mode",
		.usage = "sb",
		.help =
                "Stop all decoders but leave connection to           \n"
                "para_server open.                                   \n"
	},
	{
		.name = "stat",
		.handler = com_stat,
		.description = "print status information",
		.usage = "stat",
		.help =
		"Add para_audiod status information to para_server's \n"
		"status information and dump everything to stdout.\n"
	},
	{
		.name = "term",
		.handler = com_term,
		.description = "terminate audiod",
		.usage = "term",
		.help =
                "Stop all decoders, shut down connection to          \n"
                "para_server and exit.                               \n"
	},
	{
		.name = NULL,
	}
};

#define MAX_STREAM_SLOTS 5
static struct stream_io siol[MAX_STREAM_SLOTS];
static int playing;
static char *af_status; /* the audio format announced in server status */
static FILE *logfile;
static int audiod_socket = -1;
static struct sockaddr_un unix_addr;
static char *socket_name, *hostname;

#define AUDIOD_OFF 1
#define AUDIOD_ON 2
#define AUDIOD_STANDBY 3

static int audiod_status = AUDIOD_ON;

/*
 * log function. first argument is loglevel.
 */
void para_log(int ll, char* fmt,...)
{
	va_list argp;
	FILE *outfd;
	struct tm *tm;
	time_t t1;
	char str[MAXLINE] = "";

	if (ll < conf.loglevel_arg)
		return;
	if (!logfile && conf.logfile_given)
		logfile = open_log(conf.logfile_arg);
	if (!logfile && conf.daemon_given)
		return;
	if (!logfile) {
		if (ll < WARNING)
			outfd = stdout;
		else
			outfd = stderr;
	} else
		outfd = logfile;
	time(&t1);
	tm = localtime(&t1);
	strftime(str, MAXLINE, "%b %d %H:%M:%S", tm);
	fprintf(outfd, "%s %s ", str, hostname);
	if (conf.loglevel_arg <= INFO)
		fprintf(outfd, "%i ", ll);
	va_start(argp, fmt);
	vfprintf(outfd, fmt, argp);
	va_end(argp);
}

static char **ring_buffer;
static int rb_size = 30;
static int rb_bytes;
long long unsigned rb_serial;

static void init_ring_buffer(void)
{
	ring_buffer = para_malloc(rb_size * sizeof(char *));
	memset(ring_buffer, 0, rb_size * sizeof(char*));
}

static void ring_buffer_free(void)
{
	int i;
	for (i = 0; i < rb_size; i++) {
		if (ring_buffer[i]) {
			free(ring_buffer[i]);
			ring_buffer[i] = NULL;
		}
	}
}

static void ring_buffer_add(char *line)
{
	unsigned int p = rb_serial % rb_size;

	if (!line)
		return;
	if (ring_buffer[p]) {
		rb_bytes -= strlen(ring_buffer[p]);
		free(ring_buffer[p]);
	}
	ring_buffer[p] = para_strdup(line);
	rb_bytes += strlen(line);
	rb_serial++;
}

char *get_time_string(struct timeval *newest_stime)
{
	struct timeval now, diff, adj_stream_start, tmp;
	int total = 0, use_server_time = 1;

	if (!playing)
		return para_strdup("");
	if (audiod_status == AUDIOD_OFF)
		goto out;
	if (sa_time_diff_sign > 0)
		tv_diff(&server_stream_start, &sa_time_diff,
			&adj_stream_start);
	else
		tv_add(&server_stream_start, &sa_time_diff,
			&adj_stream_start);
	tmp = adj_stream_start;
	if (newest_stime && audiod_status == AUDIOD_ON) {
		tv_diff(newest_stime, &adj_stream_start, &diff);
		if (tv2ms(&diff) < 5000) {
			tmp = *newest_stime;
			use_server_time = 0;
		}
	}
	gettimeofday(&now, NULL);
	tv_diff(&now, &tmp, &diff);
	total = diff.tv_sec + offset_seconds;
	if (total > length_seconds)
		total = length_seconds;
	if (total < 0)
		total = 0;
out:
	return make_message(
		"%s:%s%d:%02d [%d:%02d] (%d%%/%d:%02d)",
		status_item_list[SI_PLAY_TIME],
		use_server_time? "~" : "",
		total / 60,
		total % 60,
		(length_seconds - total) / 60,
		(length_seconds - total) % 60,
		length_seconds? (total * 100 + length_seconds / 2) / length_seconds : 0,
		length_seconds / 60,
		length_seconds % 60
	);
}


static char *audiod_status_string(void)
{
	int i;
	struct timeval *newest_stime = NULL;
	char *ret, *time_string, *uptime_string, *decoder_flags =
		para_malloc((MAX_STREAM_SLOTS + 1) * sizeof(char));

	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		struct stream_io *s = &siol[i];
		char flag = '0';
		if (s->rpid > 0)
			flag += 1;
		if (s->wpid > 0)
			flag += 2;
		if (flag != '0')
			flag += s->format * 4;
		decoder_flags[i] = flag;
		if (s->wpid <= 0)
			continue;
		if (newest_stime && tv_diff(&s->wstime, newest_stime, NULL) <= 0)
			continue;
		newest_stime = &s->wstime;
	}
	decoder_flags[MAX_STREAM_SLOTS] = '\0';
	time_string = get_time_string(newest_stime);
	uptime_string = uptime_str(server_uptime(GET));
	ret = make_message("%s:%s\n%s:%s\n%s:%s\n%s",
		status_item_list[SI_AUDIOD_UPTIME], uptime_string,
		status_item_list[SI_DECODER_FLAGS], decoder_flags,
		status_item_list[SI_AUDIOD_STATUS], audiod_status == AUDIOD_ON?
			"on" : (audiod_status == AUDIOD_OFF? "off": "sb"),
		 time_string);
	free(uptime_string);
	free(decoder_flags);
	free(time_string);
	return ret;
}

static int ring_buffer_dump(int fd)
{
	unsigned int p = rb_serial % rb_size;
	char *tmp, *buf = para_strdup(NULL);
	int ret;
	ssize_t len;

	do {
		if (ring_buffer[p]) {
			tmp = buf;
			buf = make_message("%s%s\n", tmp, ring_buffer[p]);
			free(tmp);
		}
		p = (p + 1) % rb_size;
	} while (p != (rb_serial % rb_size));
	tmp = audiod_status_string();
	buf = para_strcat(buf, tmp);
	buf = para_strcat(buf, "\n");
	free(tmp);
	len = strlen(buf);
	ret = send_buffer(fd, buf);
//	PARA_DEBUG_LOG("%s: wrote %d/%d\n", __func__, ret, len);
	free(buf);
	return ret;
}

static char *configfile_exists(void)
{
	static char *config_file;

	if (!config_file) {
		char *home = para_homedir();
		config_file = make_message("%s/.paraslash/audiod.conf", home);
		free(home);
	}
	return file_exists(config_file)? config_file : NULL;
}

static void setup_signal_handling(void)
{
	if (para_signal_init(signal_pipe) < 0)
		exit(EXIT_FAILURE);
	PARA_INFO_LOG("signal pipe: fd %d\n", signal_pipe[0]);
	para_install_sighandler(SIGINT);
	para_install_sighandler(SIGTERM);
	para_install_sighandler(SIGCHLD);
	para_install_sighandler(SIGHUP);
	signal(SIGPIPE, SIG_IGN);
}

static void audiod_status_dump(void)
{
	static char *prev_status;
	char *tmp = audiod_status_string();

	if (!prev_status || strcmp(tmp, prev_status))
		num_stat_clients = dump_stat_line(tmp);
	free(prev_status);
	prev_status = tmp;
}

static long unsigned penalty_to_ms(int penalty)
{
//	return 1 << (penalty + 6);
	return MAX_PENALTY_MS * penalty * penalty / MAX_PENALTY / MAX_PENALTY;
}

static void clear_slot(int slot)
{
	struct stream_io *s = &siol[slot];
	int i;

	PARA_INFO_LOG("clearing slot %d\n", slot);
	free(s->buf);
	s->rpid = 0;
	s->wpid = 0;
	s->read_fd = 0;
	s->write_fd = 0;
	s->format = -1;
	s->rkilled = 0;
	s->wkilled = 0;
	if (!s->filters)
		return;
	for (i = 0; i < afi[s->format].num_filters; i++)
		free(s->filters[i].buf);
	free(s->filters);
	s->filters = NULL;
}

static void kill_stream_writer(int slot)
{
	struct stream_io *s = &siol[slot];

	if (s->format < 0 || s->wkilled || s->wpid <= 0)
		return;
	PARA_DEBUG_LOG("stopping %s stream writer (slot %d)\n",
		afi[s->format].name, slot);
	kill(s->wpid, SIGTERM);
	s->wkilled = 1;
	s->buflen = 0;
}

static void kill_stream_reader(int slot)
{
	struct stream_io *s = &siol[slot];

	if (s->format < 0 || s->rkilled || s->rpid <= 0)
		return;
	PARA_INFO_LOG("stopping %s stream reader (slot %d)\n",
		afi[s->format].name, slot);
	s->rkilled = 1;
	kill(s->rpid, SIGTERM);
}
static void kill_filters(int slot)
{
	int i;
	struct stream_io *s = &siol[slot];

	if (!s->filters || s->format < 0)
		return;
	for (i = 0; i < afi[s->format].num_filters; i++) {
		struct filter *f = &s->filters[i];

		if (f->pid <= 0 || f->killed)
			continue;
		PARA_INFO_LOG("slot %d: stopping %s filter %d\n",
			slot, afi[s->format].name, i);
		kill(f->pid, SIGTERM);
		f->killed = 1;
	};
}

/*
 * Send SIGTERM to reader and writer if present
 */
static void kill_decoder(int slot)
{
	kill_stream_writer(slot);
	kill_filters(slot);
	kill_stream_reader(slot);
}

static void kill_all_decoders(void)
{
	int i;

//	PARA_DEBUG_LOG("%s: stopping all decoders\n", __func__);
	for (i = 0; i < MAX_STREAM_SLOTS; i++)
		if (siol[i].format >= 0)
			kill_decoder(i);
}

static void recalc_penalty(int format, int lifetime)
{
	if (lifetime > PENALTY_RESET_TIME) {
		afi[format].penalty = 0;
		return;
	}
	if (lifetime > PENALTY_INCREASE_TIME)
		return;
	if (afi[format].penalty >= MAX_PENALTY)
		return;
	afi[format].penalty++;
	PARA_NOTICE_LOG("increasing %s penalty to %lu/%i (%lu/%i ms)\n",
		afi[format].name, afi[format].penalty,
		MAX_PENALTY,
		penalty_to_ms(afi[format].penalty),
		MAX_PENALTY_MS
	);
}

static void check_sigchld(void)
{
	pid_t pid;
	int i;
reap_next_child:
//	PARA_DEBUG_LOG("reaping\n");
	pid = para_reap_child();
	if (pid <= 0)
		return;
	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		struct timeval now;
		struct stream_io *s = &siol[i];
		long lifetime;
		int j;

		if (s->format < 0)
			continue;
		gettimeofday(&now, NULL);
		if (pid == s->rpid) {
			s->rpid = -1;
			lifetime = now.tv_sec - s->rstime.tv_sec;
			PARA_INFO_LOG("%s stream reader in slot %d died "
				"after %li secs\n",
				afi[s->format].name,
				i,
				lifetime);
			recalc_penalty(s->format, lifetime);
			break;
		}
		if (pid == s->wpid) {
			s->wpid = -1;
			lifetime = now.tv_sec - s->wstime.tv_sec;
			PARA_INFO_LOG("%s stream writer in slot %d died "
				"after %li secs\n",
				afi[s->format].name, i, lifetime);
			recalc_penalty(s->format, lifetime);
			break;
		}
		if (!s->filters)
			continue;
		for (j = 0; j < afi[s->format].num_filters; j++) {
			if (pid != s->filters[j].pid)
				continue;
			PARA_INFO_LOG("%s filter #%d in slot %d died\n",
				afi[s->format].name, j, i);
			s->filters[j].pid = -1;
			goto reap_next_child;
		}
	}
	if (i >= MAX_STREAM_SLOTS)
		PARA_CRIT_LOG("an unknown soldier #%d died\n", pid);
	goto reap_next_child;
}

static int get_empty_slot(void)
{
	int i, j;
	struct stream_io *s;

	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		s = &siol[i];
		if (s->format < 0)
			return i;
		if (s->read_fd > 0 || s->write_fd > 0 || s->rpid > 0
				|| s->wpid > 0)
			continue;
		for (j = 0; j < afi[s->format].num_filters; j++)
			if (s->filters[j].pid > 0)
				break;
		if (j < afi[s->format].num_filters)
			continue;
		clear_slot(i);
		return i;
	}
	return -1;
}

static int decoder_running(int format)
{
	int i, ret = 0;
	struct stream_io *s;

	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		s = &siol[i];
		if (s->format == format && s->rpid > 0)
			ret |= 1;
		if (s->format == format && s->wpid > 0)
			ret |= 2;
	}
	return ret;
}

static void close_stat_pipe(void)
{
	char *msg;
	if (stat_pipe < 0)
		return;
	PARA_NOTICE_LOG("%s", "closing status pipe\n");
	close(stat_pipe);
	kill_all_decoders();
	stat_pipe = -1;
	ring_buffer_free();
	dump_empty_status();
	length_seconds = 0;
	offset_seconds = 0;
	audiod_status_dump();
	playing = 0;
	msg = make_message("%s:no connection to para_server\n",
		status_item_list[SI_STATUS_BAR]);
	ring_buffer_add(msg);
	dump_stat_line(msg);
	free(msg);
}

static void clean_exit(int status)
{
	kill_all_decoders();
	if (socket_name)
		unlink(socket_name);
	if (stat_pipe >= 0)
		close_stat_pipe();
	exit(status);
}

static char *glob_cmd(char *cmd)
{
	char *ret, *replacement;
	struct timeval tmp, delay, rss; /* real stream start */

	delay.tv_sec = 0;
	delay.tv_usec = conf.stream_delay_arg;
	if (sa_time_diff_sign < 0)
		tv_add(&server_stream_start, &sa_time_diff, &rss);
	else
		tv_diff(&server_stream_start, &sa_time_diff, &rss);
	tv_add(&rss, &delay, &tmp);
	replacement = make_message("%lu:%lu", tmp.tv_sec, tmp.tv_usec);
	ret = s_a_r(cmd, "STREAM_START", replacement);
	PARA_INFO_LOG("cmd: %s, repl: %s, globbed cmd: %s\n", cmd, replacement, ret);
	free(replacement);
	return ret;
}

static int start_filters(int slot)
{
	struct stream_io *s = &siol[slot];
	struct audio_format_info *a = &afi[s->format];
	int nf = a->num_filters;
	int i;

	s->filters = para_malloc(nf * sizeof(struct filter));
	for (i = 0; i < nf; i++) {
		int fd[3] = {1, 1, 0};
		struct filter *f = &s->filters[i];
		f->buf = para_malloc(STREAM_IO_BUFSIZE);
		PARA_INFO_LOG("starting filter: %s\n", a->filter_cmds[i]);
		para_exec_cmdline_pid(&f->pid, a->filter_cmds[i], fd);
		f->write_fd = fd[0];
		f->read_fd = fd[1];
		f->loaded = 0;
		f->killed = 0;
		PARA_NOTICE_LOG("%s filter #%d started. read_fd: %d, write_fd: %d, pid: %d\n",
			a->name, i, f->read_fd, f->write_fd, f->pid);
	}
	return 1;
}

static void start_stream_writer(int slot)
{
	int ret, fds[3] = {1, 0, 0};
	struct stream_io *s = &siol[slot];
	struct audio_format_info *a = &afi[s->format];
	char *glob = glob_cmd(a->wcmd);

	start_filters(slot);
	PARA_INFO_LOG("starting stream writer: %s\n", glob? glob : a->wcmd);
	ret = para_exec_cmdline_pid(&s->wpid, glob? glob : a->wcmd, fds);
	free(glob);
	if (ret < 0) {
		PARA_ERROR_LOG("exec failed (%d)\n", ret);
		return;
	}
	s->write_fd = fds[0];
	/* we write to this fd in do_select, so we need non-blocking */
	fcntl(s->write_fd, F_SETFL, O_NONBLOCK);
	gettimeofday(&s->wstime, NULL);
	current_decoder = slot;
	PARA_NOTICE_LOG("started %s, pid %d, fd %d, slot %d\n",
		a->name, s->wpid, s->write_fd, slot);
}

static void start_stream_reader(int format)
{
	struct audio_format_info *a = &afi[format];
	struct stream_io *s;
	int ret, slot;
	int fds[3] = {0, 1, 0};

	PARA_INFO_LOG("starting stream reader: %s\n", a->rcmd);
	slot = get_empty_slot();
	if (slot < 0) {
		PARA_EMERG_LOG("%s", "no more empty slots!\n");
		clean_exit(EXIT_FAILURE);
	}
	s = &siol[slot];
	ret = para_exec_cmdline_pid(&s->rpid, a->rcmd, fds);
	if (ret < 0) {
		PARA_ERROR_LOG("exec failed (%d)\n", ret);
		return;
	}
	s->read_fd = fds[1];
	s->buf = para_malloc(a->bufsize * sizeof(char));
	s->buflen = 0;
	s->format = format;
	s->rkilled = 0;
	s->wkilled = 0;
	gettimeofday(&s->rtime, NULL);
	s->wtime = s->rtime;
	s->rstime = s->rtime;
	a->stime = s->rtime;
	PARA_NOTICE_LOG("started %s, pid %d, fd %d slot %d\n",
		a->name, s->rpid, s->read_fd, slot);
}

static int is_frozen(int format)
{
	long unsigned difftime; /* milliseconds */
	struct timeval now, tmp;
	struct audio_format_info *a = &afi[format];

	gettimeofday(&now, NULL);
	tv_diff(&now, &a->stime, &tmp);
	difftime = tv2ms(&tmp);
	return penalty_to_ms(a->penalty) > difftime? 1 : 0;
}

static void start_current_decoder(void)
{
	int i;
	struct audio_format_info *a;

	if (!af_status)
		return;
	for (i = 0, a = afi; a->name; a++, i++)
		if (!strcmp(af_status, a->name))
                        break;
	if (!a->name)
		return;
	if ((decoder_running(i) & 1) || is_frozen(i))
		return;
	start_stream_reader(i);
}

#define CONVEXITY_COEFF 5


/*
 * print status line if line starts with known command.
 */
static void check_stat_line(char *line)
{
	size_t len, ilen;
	const char *s;
	struct timeval max_deviation = {0, 500 * 1000};


	if (!line)
		return;
	len = strlen(line);

	ring_buffer_add(line);
	num_stat_clients = dump_stat_line(line);

	s = status_item_list[SI_STATUS];
	ilen = strlen(s);
	if (!strncmp(line, s, ilen) && line[ilen] == ':') {
		if (len > ilen + 7 && !strncmp(line + ilen, ":playing", 8))
			playing = 1;
		else
			playing = 0;
		return;
	}
	s = status_item_list[SI_FORMAT];
	ilen = strlen(s);
	if (!strncmp(line, s, ilen) && line[ilen] == ':') {
		free(af_status);
		af_status = para_strdup(line + ilen + 1);
		return;
	}
	s = status_item_list[SI_OFFSET];
	ilen = strlen(s);
	if (!strncmp(line, s, ilen) && line[ilen] == ':') {
		offset_seconds = atoi(line + ilen + 1);
		return;
	}
	s = status_item_list[SI_LENGTH];
	ilen = strlen(s);
	if (!strncmp(line, s, ilen) && line[ilen] == ':') {
		length_seconds = atoi(line + ilen + 1);
		return;
	}
	s = status_item_list[SI_STREAM_START];
	ilen = strlen(s);
	if (!strncmp(line, s, ilen) && line[ilen] == ':') {
		struct timeval tmp;
		if (sscanf(line + ilen + 1, "%lu.%lu", &tmp.tv_sec, &tmp.tv_usec) == 2)
			server_stream_start = tmp;
		return;
	}
	s = status_item_list[SI_CURRENT_TIME];
	ilen = strlen(s);
	if (!strncmp(line, s, ilen) && line[ilen] == ':') {
		struct timeval st, now, tmp, diff;
		static int count;
		int sign;
		if (sscanf(line + ilen + 1, "%lu.%lu", &st.tv_sec, &st.tv_usec) != 2)
			return;
		gettimeofday(&now, NULL);
		sign = tv_diff(&st, &now, &diff);
//		PARA_NOTICE_LOG("%s: sign = %i, sa_time_diff_sign = %i\n", __func__,
//			sign, sa_time_diff_sign);
		if (!count) {
			sa_time_diff_sign = sign;
			sa_time_diff = diff;
			count++;
			return;
		}
		if (count > 5) {
			int s = tv_diff(&diff, &sa_time_diff, &tmp);
			if (tv_diff(&max_deviation, &tmp, NULL) < 0)
				PARA_WARNING_LOG("time diff jump: %lims\n",
					s * tv2ms(&tmp));
		}
		count++;
		sa_time_diff_sign = tv_convex_combination(
			sa_time_diff_sign * CONVEXITY_COEFF, &sa_time_diff,
			count > 10? sign : sign * CONVEXITY_COEFF, &diff,
			&tmp);
		sa_time_diff = tmp;
		PARA_INFO_LOG("time diff (cur/avg): "
			"%li:%lu/%li:%lu\n",
			sign * diff.tv_sec, (diff.tv_usec + 500) / 1000,
			sa_time_diff_sign * sa_time_diff.tv_sec,
			(sa_time_diff.tv_usec + 500)/ 1000
		);
	}
}

static void handle_signal(int sig)
{
	switch (sig) {
	case SIGCHLD:
		return check_sigchld();
	case SIGINT:
	case SIGTERM:
	case SIGHUP:
		PARA_EMERG_LOG("terminating on signal %d\n", sig);
		clean_exit(EXIT_FAILURE);
		return;
	}
}

static void check_timeouts(void)
{
	struct timeval now;
	int slot, timeout = conf.stream_timeout_arg;

	gettimeofday(&now, NULL);
	for (slot = 0; slot < MAX_STREAM_SLOTS; slot++) {
		struct stream_io *s = &siol[slot];
		if (s->format < 0)
			continue;
		/* check read time */
		if (s->rpid > 0 && !s->rkilled &&
			now.tv_sec > s->rtime.tv_sec + timeout) {
			PARA_DEBUG_LOG("%s input buffer (slot %d) not ready\n",
				afi[s->format].name, slot);
			kill_stream_reader(slot);
		}
		/* check write time */
		if (s->wpid > 0 && !s->wkilled &&
			now.tv_sec > s->wtime.tv_sec + timeout) {
			PARA_DEBUG_LOG("%s output buffer (slot %d) not ready\n",
				afi[s->format].name, slot);
			kill_stream_writer(slot);
		}
	}
}

static void close_unused_stream_fds(int slot)
{
	struct stream_io *s = &siol[slot];
	struct audio_format_info *a;
	pid_t pid;
	unsigned int loaded, nf;
	int i, other_fd;

	if (s->format < 0)
		return;
	a = &afi[s->format];
	nf = a->num_filters;
	/* writer */
	if (s->write_fd <= 0)
		goto close_filters;
	if (nf && !s->filters)
		goto close_reader;
	pid = s->wpid;
	other_fd = a->num_filters? s->filters[nf - 1].read_fd : s->read_fd;
	loaded = s->buflen;
	if (pid <= 0 || (other_fd < 0 && !loaded)) {
		PARA_INFO_LOG("slot %d: closing write fd %d\n",
			slot, s->write_fd);
		close(s->write_fd);
		s->write_fd = -1;
	}
close_filters:
	if (!s->filters)
		goto close_reader;
	for (i = 0; i < nf; i++) {
		struct filter *f = &s->filters[i];

		pid = f->pid;
		if (f->write_fd <= 0)
			goto close_filter_read_fd;
		/* write fd */
		other_fd = i? s->filters[i - 1].read_fd : s->read_fd;
		loaded = i? s->filters[i - 1].loaded : s->buflen;
		if (pid <= 0 || (other_fd < 0 && !loaded)) {
			PARA_INFO_LOG("slot %d: closing write fd of %s filter %d, fd %d\n",
				slot, a->name, i, f->write_fd);
			close(f->write_fd);
			f->write_fd = -1;
		}
close_filter_read_fd:
		if (f->read_fd <= 0)
			continue;
		other_fd = (i == nf - 1)? s->write_fd : s->filters[i + 1].write_fd;
//		PARA_INFO_LOG("slot %d, filter %d, pid: %d, other fd: %d\n", slot, i, pid, other_fd);
		if (pid <= 0 || other_fd < 0) {
			PARA_INFO_LOG("slot %d: closing read fd of %s filter %d, fd %d\n",
				slot, a->name, i, f->read_fd);
			close(f->read_fd);
			f->read_fd = -1;
		}
	}
close_reader:
	if (s->read_fd <= 0)
		return;
	if (nf && !s->filters)
		return;
	pid = s->rpid;
	other_fd = nf? s->filters[0].write_fd : s->write_fd;
	if (pid < 0 || other_fd < 0) {
//	if (s->write_fd < 0 || s->rpid < 0) {
		PARA_INFO_LOG("slot %d: closing read fd %d\n",
			slot, s->read_fd);
		close(s->read_fd);
		s->read_fd = -1;
		s->buflen = 0;
	}
}

static int set_filter_fds(int slot, fd_set *rfds, fd_set *wfds)
{
	struct stream_io *s = &siol[slot];
	struct audio_format_info *a;
	int i, max_fileno = -1;

	if (s->format < 0 || !s->filters)
		return -1;
	a = &afi[s->format];
//	PARA_DEBUG_LOG("format: %d. name: num_filters: %d, %s\n", s->format, a->num_filters, a->name);
	for (i = 0; i < a->num_filters; i++) {
		struct filter *f = &s->filters[i];
		unsigned int loaded = i? s->filters[i - 1].loaded : s->buflen;

		if (loaded > 0 && f->write_fd > 0) {
//			PARA_DEBUG_LOG("FD_SETting write_fd %d for filter #%d, loaded: %d\n", f->write_fd, i, loaded);
			FD_SET(f->write_fd, wfds);
			f->wcheck = 1;
			max_fileno = MAX(max_fileno, f->write_fd);
		}
		if (f->read_fd > 0 && f->loaded < STREAM_IO_BUFSIZE) {
//			PARA_DEBUG_LOG("FD_SETting read_fd %d for filter #%d\n", f->read_fd, i);
			FD_SET(f->read_fd, rfds);
			f->rcheck = 1;
			max_fileno = MAX(max_fileno, f->read_fd);
		}
	}
//	PARA_DEBUG_LOG("returning %d\n", max_fileno);
	return max_fileno;
}

static int set_stream_fds(fd_set *rfds, fd_set *wfds)
{
	int i, max_fileno = -1;

//	PARA_INFO_LOG("%s", "start\n");
	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		struct stream_io *s = &siol[i];
		struct audio_format_info *a;

		s->rcheck = 0;
		s->wcheck = 0;
		if (s->format < 0)
			continue;
		a = &afi[s->format];
//		PARA_INFO_LOG("slot %d: closing unused fds\n", i);
		close_unused_stream_fds(i);
//		PARA_INFO_LOG("slot %d: closed unused fds\n", i);
		if (s->read_fd > 0 && s->buflen < afi[s->format].bufsize) {
//			PARA_DEBUG_LOG("slot %d: buflen = %d, s->read = %d "
//				"FDSET for read\n",
//				i, s->buflen, s->read_fd);
			FD_SET(s->read_fd, rfds);
			s->rcheck = 1;
			max_fileno = MAX(s->read_fd, max_fileno);
		}
		if (s->buflen && !s->wpid) {
			PARA_INFO_LOG("no writer in slot %d\n", i);
			start_stream_writer(i);
		}

		max_fileno = MAX(max_fileno, set_filter_fds(i, rfds, wfds));
		if (s->write_fd <= 0)
			continue;
		if ((!a->num_filters && !s->buflen) || (a->num_filters && s->filters && !s->filters[a->num_filters - 1].loaded))
			continue;
//		PARA_DEBUG_LOG("slot %d: buflen = %d, " "s->write = %d DFSET for write\n", i, s->buflen, s->write_fd);
			/*
			 * buf already contains data which was not yet
			 * witten
			 */
		FD_SET(s->write_fd, wfds);
		s->wcheck = 1;
		max_fileno = MAX(s->write_fd, max_fileno);
	}
//	PARA_INFO_LOG("return %d\n", max_fileno);
	return max_fileno;
}

static int write_audio_data(int slot)
{
	struct stream_io *s = &siol[slot];
	struct audio_format_info *a = &afi[s->format];
	int rv;
	char **buf;
	unsigned int *len;

//	PARA_INFO_LOG("slot %d\n", slot);
	if (a->num_filters) {
		buf = &s->filters[a->num_filters - 1].buf;
		len = &(s->filters[a->num_filters - 1].loaded);
	} else {
		buf = &s->buf;
		len = &s->buflen;
	}
	rv = write(s->write_fd, *buf, *len);
	PARA_DEBUG_LOG("wrote %d\n", rv);
	if (rv < 0) {
		PARA_WARNING_LOG("write error in slot %d (fd %d): %s\n",
			 slot, s->write_fd, strerror(errno));
		kill_decoder(slot);
		*len = 0;
	} else if (rv != *len) {
		PARA_DEBUG_LOG("partial %s write (%i/%i) for slot %d\n",
			a->name, rv, *len, slot);
		*len -= rv;
		memmove(*buf, *buf + rv, *len);
	} else {
		gettimeofday(&s->wtime, NULL);
		*len = 0;
	}
	return rv;
}

static int read_audio_data(int slot)
{
	struct stream_io *s = &siol[slot];
	struct audio_format_info *a = &afi[s->format];
	int ret;

//	PARA_DEBUG_LOG("reading at most %d byte from %s pipe (fileno %d)\n",
//			a->bufsize, a->name, file_no);
	ret = read(s->read_fd, s->buf + s->buflen, a->bufsize - s->buflen);
	PARA_DEBUG_LOG("read %d\n", ret);
	if (ret > 0)
		s->buflen += ret;
	if (s->wpid < 0)
		s->buflen = 0;
	if (ret < 0) {
		PARA_ERROR_LOG("%s read error (%s)\n",
			a->name, strerror(errno));
		kill_decoder(slot);
		s->buflen = 0;
	} else if (ret > 0)
		gettimeofday(&s->rtime, NULL);
	else {
		PARA_INFO_LOG("eof in slot %d, closing read fd %d "
			"(%d bytes left)\n", slot, s->read_fd,
			s->buflen);
		close(s->read_fd);
		s->read_fd = -1;
	}
//	PARA_DEBUG_LOG("slot %d: buflen = %d\n", slot, s->buflen);
	return s->buflen;
}

static void do_filter_io(int slot, fd_set *rfds, fd_set *wfds)
{
	struct stream_io *s = &siol[slot];
	struct audio_format_info *a;
	unsigned int nf;
	int i, ret;

	if (s->format < 0 || !s->filters)
		return;
	a = &afi[s->format];
	nf = a->num_filters;
	for (i = 0; i < nf; i++) {
		struct filter *f = &s->filters[i];
		char **buf;
		unsigned int *len;

		if (!f->rcheck || f->read_fd < 0 || !FD_ISSET(f->read_fd, rfds))
			goto check_write;
		ret = read(f->read_fd, f->buf + f->loaded, STREAM_IO_BUFSIZE - f->loaded);
//		PARA_DEBUG_LOG("slot %d, filter %d: read %d\n", slot, i, ret);
		if (ret <= 0) {
			PARA_DEBUG_LOG("slot %d: closing read fd of filter %d\n", slot, i);
			close(f->read_fd);
			f->read_fd = -1;
		}
		f->loaded += ret;
check_write:
		PARA_DEBUG_LOG("slot %d filter %i, fd %d\n", slot, i, f->write_fd);
		if (!f->wcheck || f->write_fd < 0 || !FD_ISSET(f->write_fd, wfds))
			continue;
		if (i) {
			buf = &s->filters[i - 1].buf;
			len = &(s->filters[i - 1].loaded);
		} else {
			buf = &s->buf;
			len = &s->buflen;
		}
		ret = write(f->write_fd, *buf, *len);
		PARA_DEBUG_LOG("slot %d, filter #%d: wrote %d\n", slot, i, ret);
		if (ret < 0) {
			PARA_DEBUG_LOG("slot %d: closing write fd of filter %d\n", slot, i);
			close(f->write_fd);
			f->write_fd = -1;
			continue;
		}
		*len -= ret;
		if (ret == *len)
			continue;
		memmove(*buf, *buf + ret, *len);
	}
}

static void do_stream_io(fd_set *rfds, fd_set *wfds)
{
	int i;

//	PARA_INFO_LOG("%s\n", "start");
	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		struct stream_io *s = &siol[i];
		if (s->rcheck && FD_ISSET(s->read_fd, rfds))
			read_audio_data(i);
		do_filter_io(i, rfds, wfds);
		if (s->wcheck && FD_ISSET(s->write_fd, wfds))
			write_audio_data(i);
	}
//	PARA_INFO_LOG("%s\n", "end");
}

static int parse_stream_command(const char *txt, char **cmd)
{
	int ret = -1;
	char *p = strchr(txt, ':');

	if (!p)
		goto out;
	p++;
	for (ret = 0; afi[ret].name; ret++) {
		if (strncmp(txt, afi[ret].name, strlen(afi[ret].name)))
			continue;
		*cmd = p;
		goto out;
	}
	ret = -2;
out:
	return ret;
}

static int init_stream_io(void)
{
	const char errmsg[] = "invalid audio format given in ";
	int i, ret = -1;
	char *cmd;

	for (i = 0; i < conf.stream_read_cmd_given; i++) {
		int f = parse_stream_command(conf.stream_read_cmd_arg[i], &cmd);
		if (f < 0) {
			PARA_ERROR_LOG("%s%s\n", errmsg, "stream_read_cmd");
			goto out;
		}
		afi[f].rcmd = para_strdup(cmd);
		PARA_INFO_LOG("%s read command: %s\n", afi[f].name, afi[f].rcmd);
	}
	for (i = 0; i < conf.filter_cmd_given; i++) {
		int nf, f = parse_stream_command(conf.filter_cmd_arg[i], &cmd);
		if (f < 0) {
			PARA_ERROR_LOG("%s%s\n", errmsg, "filter");
			goto out;
		}
		nf = afi[f].num_filters++;
		afi[f].filter_cmds = para_realloc(afi[f].filter_cmds, (nf + 1)
			* sizeof(char *));
		afi[f].filter_cmds[nf] = para_strdup(cmd);
		PARA_INFO_LOG("%s filter #%d: %s\n", afi[f].name, nf, afi[f].filter_cmds[nf]);
	}
	for (i = 0; i < conf.stream_write_cmd_given; i++) {
		int f = parse_stream_command(conf.stream_write_cmd_arg[i], &cmd);
		if (f < 0) {
			PARA_ERROR_LOG("%s%s\n", errmsg, "stream_write_cmd");
			goto out;
		}
		afi[f].wcmd = para_strdup(cmd);
		PARA_INFO_LOG("%s write command: %s\n", afi[f].name, afi[f].wcmd);
	}
	ret = 1;
out:
	return ret;
}

static int dump_commands(int fd)
{
	char *buf = para_strdup(""), *tmp = NULL;
	int i;
	ssize_t ret;

	for (i = 0; cmds[i].name; i++) {
		tmp = make_message("%s%s\t%s\n", buf, cmds[i].name,
			cmds[i].description);
		free(buf);
		buf = tmp;
	}
	ret = write(fd, buf, strlen(buf));
	free(buf);
	return ret;
}

int com_help(int fd, int argc, char **argv)
{
	int i, ret = -1;
	char *buf;
	const char *dflt = "No such command. Available commands:\n";

	if (argc < 2) {
		ret = dump_commands(fd);
		goto out;
	}
	for (i = 0; cmds[i].name; i++) {
		if (strcmp(cmds[i].name, argv[0]))
			continue;
		buf = make_message(
			"NAME:\n~~~~~\n\t%s -- %s\n\n"
			"USAGE:\n~~~~~~\n\tpara_audioc %s\n\n"
			"DESCRIPTION:\n~~~~~~~~~~~~\n%s\n",
			argv[0],
			cmds[i].description,
			cmds[i].usage,
			cmds[i].help
		);
		ret = write(fd, buf, strlen(buf));
		free(buf);
		goto out;
	}
	ret = write(fd, dflt, strlen(dflt));
	if (ret > 0)
		ret = dump_commands(fd);
out:
	close(fd);
	return ret;
}

int com_stat(int fd, int argc, char **argv)
{
	int ret = -1;

	if (num_stat_clients > MAX_STAT_CLIENTS) {
		PARA_ERROR_LOG("maximal number of stat clients (%d) exceeded\n",
			MAX_STAT_CLIENTS);
		goto out;
	}
	ret = ring_buffer_dump(fd);
	if (ret > 0)
		num_stat_clients = stat_client_add(fd);
out:
	if (ret < 0)
		close(fd);
	return ret;
}

int com_term(int fd, int argc, char **argv)
{
	kill_all_decoders();
	PARA_NOTICE_LOG("%s", "terminating on user request\n");
	close(fd);
	clean_exit(EXIT_SUCCESS);
	return 42;
}

int com_on(int fd, int argc, char **argv)
{
	audiod_status = AUDIOD_ON;
	close(fd);
	return 1;
}

int com_off(int fd, int argc, char **argv)
{
	audiod_status = AUDIOD_OFF;
	close(fd);
	return 1;
}

int com_sb(int fd, int argc, char **argv)
{
	audiod_status = AUDIOD_STANDBY;
	close(fd);
	return 1;
}

int com_cycle(int fd, int argc, char **argv)
{
	switch (audiod_status) {
		case  AUDIOD_ON:
			return com_sb(fd, argc, argv);
			break;
		case  AUDIOD_OFF:
			return com_on(fd, argc, argv);
			break;
		case  AUDIOD_STANDBY:
			return com_off(fd, argc, argv);
			break;
	}
	close(fd);
	return 1;
}

static int check_perms(struct ucred *c)
{
	int i;

	if (!conf.user_allow_given)
		return 1;
	for (i = 0; i < conf.user_allow_given; i++)
		if (c->uid == conf.user_allow_arg[i])
			return 1;
	return -1;
}

static int handle_connect(void)
{
	int i, argc, ret, clifd = -1;
	size_t len = sizeof(struct sockaddr_un);
	struct ucred c;
	char *buf = para_malloc(MAXLINE), **argv;
	/*
	 * The len argument is a value-result parameter: it should
	 * initially contain the size of the structure pointed  to
	 * by  addr;  on  return it will contain the actual length (in
	 * bytes) of the address returned.
	 */
	ret = -1;
	clifd = accept(audiod_socket, (struct sockaddr *)&unix_addr, &len);
	if (clifd < 0)
		goto out;
	ret = -2;
	len = recv_cred_buffer(clifd, buf, MAXLINE - 1, &c);
	if (len < 0)
		goto out;
	PARA_DEBUG_LOG("received: %s (%d bytes)\n", buf, len);
//	fprintf(stderr, "pid: %i, uid: %i, gid: %i\n", c.pid, c.uid, c.gid);
	ret = -3;
	if (check_perms(&c) <= 0)
		goto out;
	argc = split_args(buf, &argv, '\n');
	PARA_DEBUG_LOG("argv[0]: %s\n", argv[0]);
	for (i = 0; cmds[i].name; i++) {
		if (strcmp(cmds[i].name, argv[0]))
			continue;
		ret = cmds[i].handler(clifd, argc, ++argv);
		goto out;
	}
	ret = -4; /* cmd not found */
	close(clifd);
out:
	free(buf);
	return ret;
}

static void audiod_get_socket(void)
{
	if (conf.socket_given)
		socket_name = para_strdup(conf.socket_arg);
	else {
		char *hn = para_hostname();
		socket_name = make_message("/var/paraslash/audiod_socket.%s",
			hn);
		free(hn);
	}
	PARA_NOTICE_LOG("connecting to local socket %s\n", socket_name);
	if (conf.force_given)
		unlink(socket_name);
	audiod_socket = create_pf_socket(socket_name, &unix_addr,
			S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IWOTH);
	if (audiod_socket < 0) {
		PARA_EMERG_LOG("%s", "can not connect to socket\n");
		exit(EXIT_FAILURE); /* do not unlink socket */
	}
	if (listen(audiod_socket, 5) < 0) {
		PARA_EMERG_LOG("%s", "can not listen on socket\n");
		exit(EXIT_FAILURE); /* do not unlink socket */
	}
}

static int open_stat_pipe(void)
{
	int ret, fd[3] = {-1, 1, 0};
	char *argv[] = {BINDIR "/para_client", "stat", NULL};
	pid_t pid;
	ret = para_exec(&pid, BINDIR "/para_client",  argv, fd);
	if (ret >= 0) {
		ret = fd[1];
		PARA_NOTICE_LOG("stat pipe opened, fd %d\n", ret);
	}
	return ret;
}

static int audiod_mainloop(void)
{
	fd_set rfds, wfds;
	int ret, max_fileno, sp_numread, sbo = 0;
	char status_buf[STRINGSIZE] = "";
	struct timeval tv;
repeat:
	FD_ZERO(&rfds);
	FD_ZERO(&wfds);
	max_fileno = 0;
	check_timeouts();
	if (audiod_status != AUDIOD_ON)
		kill_all_decoders();
	else if (playing)
		start_current_decoder();
	max_fileno = set_stream_fds(&rfds, &wfds);
	/* stat pipe (read) */
	if (stat_pipe >= 0 && audiod_status == AUDIOD_OFF)
		close_stat_pipe();
	if (stat_pipe < 0 && audiod_status != AUDIOD_OFF)
		stat_pipe = open_stat_pipe();
	if (stat_pipe >= 0 && audiod_status != AUDIOD_OFF) {
		FD_SET(stat_pipe, &rfds);
		max_fileno = MAX(max_fileno, stat_pipe);
	}
	/* signal pipe */
	FD_SET(signal_pipe[0], &rfds);
	max_fileno = MAX(max_fileno, signal_pipe[0]);
	/* local socket */
	if (audiod_socket < 0)
		audiod_get_socket();
	FD_SET(audiod_socket, &rfds);
	max_fileno = MAX(max_fileno, audiod_socket);
	tv.tv_sec = 0;
	tv.tv_usec = 200000;
	ret = select(max_fileno + 1, &rfds, &wfds, NULL, &tv);

	if (ret < 0 && errno != EINTR)
		PARA_ERROR_LOG("select returned %d (%s)\n",
			ret, strerror(errno));
	if (audiod_status != AUDIOD_OFF)
		audiod_status_dump();
	if (ret <= 0)
		goto repeat;
	do_stream_io(&rfds, &wfds);
	/* read status pipe */
	if (stat_pipe >=0 && FD_ISSET(stat_pipe, &rfds)) {
		if ((sp_numread = read(stat_pipe, status_buf + sbo,
				STRINGSIZE - 1 - sbo)) <= 0) {
			close_stat_pipe();
			stat_pipe = open_stat_pipe();
			/* avoid busy loop if server is down */
			while (sleep(1) > 0)
				; /* try again*/
		}
		status_buf[sp_numread + sbo] = '\0';
		sbo = check_buf_for_items(status_buf, sp_numread + sbo,
			&check_stat_line, 0);
	}
	if (FD_ISSET(audiod_socket, &rfds)) {
		ret = handle_connect();
		if (ret <= 0)
			PARA_ERROR_LOG(" command handler returned %d\n", ret);
	}
	/* signals */
	if (FD_ISSET(signal_pipe[0], &rfds)) {
		int sig_nr = para_next_signal();
		if (sig_nr > 0)
			handle_signal(sig_nr);
	}
	goto repeat;
	return 42; /* never reached */
}

static void set_initial_status(void)
{
	audiod_status = AUDIOD_ON;
	if (!conf.mode_given)
		return;
	if (!strcmp(conf.mode_arg, "sb")) {
		audiod_status = AUDIOD_STANDBY;
		return;
	}
	if (!strcmp(conf.mode_arg, "off")) {
		audiod_status = AUDIOD_OFF;
		return;
	}
	if (strcmp(conf.mode_arg, "on"))
		PARA_WARNING_LOG("%s", "invalid mode\n");
}

int main(int argc, char *argv[])
{
	char *cf;
	int i;


	hostname = para_hostname();
	cmdline_parser(argc, argv, &conf);
	para_drop_privileges(conf.user_arg);
	cf = configfile_exists();
	if (cf) {
		if (cmdline_parser_configfile(cf, &conf, 0, 0, 0)) {
			fprintf(stderr, "parse error in config file\n");
			exit(EXIT_FAILURE);
		}
	}
	log_welcome("para_audiod", conf.loglevel_arg);
	i = init_stream_io();
	if (i < 0) {
		fprintf(stderr, "init stream io error\n");
		exit(EXIT_FAILURE);
	}
	server_uptime(SET); /* reset server uptime */
	set_initial_status();
	for (i = 0; i < MAX_STREAM_SLOTS; i++) {
		siol[i].buf = NULL;
		siol[i].filters = NULL;
		clear_slot(i);
	}
	init_ring_buffer();
	setup_signal_handling();
	if (conf.daemon_given)
		daemon_init();
	return audiod_mainloop(); /* never returns */
}
